﻿using System;
using System.Collections;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Reflection;

namespace TestProject
{
    /// <summary>
    /// 자동 필터 컬렉션
    /// </summary>
    /// <typeparam name="TItem">항목 타입</typeparam>
    public class AutoFilterCollection<TItem> : INotifyPropertyChanged
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Event
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 속성 변경시 이벤트 - PropertyChanged

        /// <summary>
        /// 속성 변경시 이벤트
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 소스 열거 가능형
        /// </summary>
        private IEnumerable<TItem> sourceEnumerable;

        /// <summary>
        /// 자동 필터 컬럼 리스트
        /// </summary>
        private readonly List<AutoFilterColumn<TItem>> autoFilterColumnList = new List<AutoFilterColumn<TItem>>();

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Property
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 소스 열거 가능형 - SourceEnumerable

        /// <summary>
        /// 소스 열거 가능형
        /// </summary>
        public IEnumerable<TItem> SourceEnumerable
        {
            get
            {
                return this.sourceEnumerable;
            }
            set
            {
                if(this.sourceEnumerable != value)
                {
                    this.sourceEnumerable = value;

                    Calculate();
                }
            }
        }

        #endregion
        #region 자동 필터 컬럼 리스트 - AutoFilterColumnList

        /// <summary>
        /// 자동 필터 컬럼 리스트
        /// </summary>
        public List<AutoFilterColumn<TItem>> AutoFilterColumnList
        {
            get
            {
                return this.autoFilterColumnList;
            }
        }

        #endregion
        #region 필터 열거 가능형 - FilteredEnumerable

        /// <summary>
        /// 필터 열거 가능형
        /// </summary>
        public IEnumerable<TItem> FilteredEnumerable
        {
            get
            {
                IEnumerable<TItem> resultEnumerable = SourceEnumerable;
                    
                foreach(AutoFilterColumn<TItem> autoFilterColumn in AutoFilterColumnList)
                {
                    if(autoFilterColumn.Value == null || autoFilterColumn.Value.Equals("All"))
                    {
                        continue;
                    }

                    PropertyInfo propertyInfo = typeof(TItem).GetProperty(autoFilterColumn.Name);

                    resultEnumerable = resultEnumerable.Where(x => propertyInfo.GetValue(x, null).Equals(autoFilterColumn.Value));
                }

                return resultEnumerable;
            }
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Internal

        #region 모두 통지하기 - NotifyAll()

        /// <summary>
        /// 모두 통지하기
        /// </summary>
        internal void NotifyAll()
        {
            foreach(AutoFilterColumn<TItem> autoFilterColumn in AutoFilterColumnList)
            {
                autoFilterColumn.Notify();
            }

            FirePropertyChangedEvent("FilteredEnumerable");
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 필터용 값 열거 가능형 구하기 - GetValueEnumerableForFilter(columnName)

        /// <summary>
        /// 필터용 값 열거 가능형 구하기
        /// </summary>
        /// <param name="columnName">컬럼명</param>
        /// <returns>필터용 값 열거 가능형</returns>
        public IEnumerable GetValueEnumerableForFilter(string columnName)
        {
            IEnumerable<TItem> resultEnumerable = SourceEnumerable;

            foreach(AutoFilterColumn<TItem> autoFilterColumn in AutoFilterColumnList)
            {
                if(autoFilterColumn.Name == columnName)
                {
                    continue;
                }

                if(autoFilterColumn.Value == null || autoFilterColumn.Value.Equals("All"))
                {
                    continue;
                }

                PropertyInfo autoFilterColumnPropertyInfo = typeof(TItem).GetProperty(autoFilterColumn.Name);

                if(autoFilterColumn.Level == null || AutoFilterColumnList.FirstOrDefault(x => x.Name == columnName).Level >= autoFilterColumn.Level)
                {
                    resultEnumerable = resultEnumerable.Where(x => autoFilterColumnPropertyInfo.GetValue(x, null).Equals(autoFilterColumn.Value));
                }
            }

            PropertyInfo propertyInfo = typeof(TItem).GetProperty(columnName);

            List<object> propertyValueList = resultEnumerable.Select(x => propertyInfo.GetValue(x, null)).Concat(new List<object>()).ToList();

            propertyValueList.Sort();

            return propertyValueList.Distinct();
        }

        #endregion
        #region 자동 필터 컬럼 구하기 - GetAutoFilterColumn(columnName)

        /// <summary>
        /// 자동 필터 컬럼 구하기
        /// </summary>
        /// <param name="columnName">컬럼명</param>
        /// <returns>자동 필터 컬럼</returns>
        public AutoFilterColumn<TItem> GetAutoFilterColumn(string columnName)
        {
            return AutoFilterColumnList.SingleOrDefault(x => x.Name == columnName);
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Protected

        #region 속성 변경시 이벤트 발생시키기 - FirePropertyChangedEvent(propertyName)

        /// <summary>
        /// 속성 변경시 이벤트 발생시키기
        /// </summary>
        /// <param name="propertyName">속성명</param>
        protected void FirePropertyChangedEvent(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }

        #endregion

        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region 계산하기 - Calculate()

        /// <summary>
        /// 계산하기
        /// </summary>
        private void Calculate()
        {
            PropertyInfo[] propertyInfoArray = typeof(TItem).GetProperties(BindingFlags.Instance | BindingFlags.Public);
                
            foreach(PropertyInfo propertyInfo in propertyInfoArray)
            {
                int? priorityLevel = null;

                object[] attributeArray = propertyInfo.GetCustomAttributes(typeof(Attribute), true);

                if(attributeArray != null && attributeArray.Length > 0)
                {
                    PriorityLevelAttribute priorityLevelAttribute = attributeArray[0] as PriorityLevelAttribute;

                    if(priorityLevelAttribute != null)
                    {
                        priorityLevel = priorityLevelAttribute.Value;
                    }
                }

                AutoFilterColumnList.Add
                (
                    new AutoFilterColumn<TItem>
                    {
                        Parent = this,
                        Level  = priorityLevel,
                        Name   = propertyInfo.Name,
                        Value  = null
                    }
                );
            }
        }

        #endregion
    }
}